@@ -2,6 +2,7 @@
 from __future__ import absolute_import
 from __future__ import unicode_literals
 
+import json
 from collections import OrderedDict
 from itertools import groupby
 
@@ -18,14 +19,15 @@ from django.utils.html import conditional_escape, format_html
 from django.utils.safestring import mark_safe
 from django.db.models import QuerySet
 from six import string_types
-from tri.declarative import declarative, creation_ordered, with_meta, setdefaults, evaluate_recursive, evaluate, getattr_path, collect_namespaces
-from tri.form import extract_subkeys, Field, Form
+from tri.declarative import declarative, creation_ordered, with_meta, setdefaults, evaluate_recursive, evaluate, getattr_path, collect_namespaces, extract_subkeys, sort_after, LAST, setdefaults_path, dispatch, EMPTY
+from tri.form import Field, Form, member_from_model, expand_member, create_members_from_model
 from tri.named_struct import NamedStructField, NamedStruct
 from tri.struct import Struct, Frozen, merged
-from tri.query import Query, Variable, QueryException
+from tri.query import Query, Variable, QueryException, Q_OP_BY_OP
 
+__version__ = '2.0.0'
 
-__version__ = '1.13.0'
+LAST = LAST
 
 
 def prepare_headers(request, bound_columns):
@@ -120,6 +122,7 @@ class ColumnBase(NamedStruct):
 
     name = NamedStructField()
     """ :type: unicode """
+    after = NamedStructField()
     attrs = NamedStructField()
     attr = NamedStructField(default=lambda table, column: column.name)
     css_class = NamedStructField(default=set())
@@ -143,7 +146,18 @@ class ColumnBase(NamedStruct):
 
 @creation_ordered
 class Column(Frozen, ColumnBase):
-    # noinspection PyShadowingBuiltins
+    @dispatch(
+        bulk__show=False,
+        query__show=False,
+        attrs=EMPTY,
+        attrs__class=EMPTY,
+        cell__template=None,
+        cell__attrs=EMPTY,
+        cell__value=lambda table, column, row: getattr_path(row, evaluate(column.attr, table=table, column=column)),
+        cell__format=default_cell_formatter,
+        cell__url=None,
+        cell__url_title=None,
+    )
     def __init__(self, **kwargs):
         """
         :param name: the name of the column
@@ -166,33 +180,12 @@ class Column(Frozen, ColumnBase):
         :param cell__url_title: callable that receives kw arguments: `table`, `column`, `row` and `value`.
         """
 
-        kwargs.update({'attrs__class__' + c: True for c in kwargs.get('css_class', {})})
+        setdefaults_path(kwargs, {'attrs__class__' + c: True for c in kwargs.get('css_class', {})})
+        super(Column, self).__init__(**kwargs)
 
-        setdefaults(kwargs, dict(
-            bulk__show=False,
-            query__show=False,
-            extra=Struct(),
-            attrs={},
-            cell__template=None,
-            cell__value=lambda table, column, row: getattr_path(row, evaluate(column.attr, table=table, column=column)),
-            cell__format=default_cell_formatter,
-            cell__attrs={},
-            cell__url=None,
-            cell__url_title=None
-        ))
-        namespaces = Struct(collect_namespaces(kwargs))
-        namespaces.attrs = Struct(collect_namespaces(namespaces.attrs))
-        namespaces.cell = Struct(collect_namespaces(namespaces.cell))
-        namespaces.cell.attrs = Struct(collect_namespaces(namespaces.cell.attrs))
-
-        namespaces.bulk = Struct(namespaces.bulk)
-        namespaces.query = Struct(namespaces.query)
-        namespaces.extra = Struct(namespaces.extra)
-
-        setdefaults(namespaces.attrs, {'class': {}})
-        setdefaults(namespaces.cell.attrs, {'class': {}})
-
-        super(Column, self).__init__(**namespaces)
+    @staticmethod
+    def text(**kwargs):
+        return Column(**kwargs)
 
     @staticmethod
     def icon(icon, is_report=False, icon_title='', show=True, **kwargs):
@@ -327,6 +320,14 @@ class Column(Frozen, ColumnBase):
         ))
         return Column(**kwargs)
 
+    @staticmethod
+    def float(**kwargs):
+        return Column.number(**kwargs)
+
+    @staticmethod
+    def integer(**kwargs):
+        return Column.number(**kwargs)
+
     @staticmethod
     def choice_queryset(**kwargs):
         setdefaults(kwargs, dict(
@@ -337,6 +338,17 @@ class Column(Frozen, ColumnBase):
         ))
         return Column.choice(**kwargs)
 
+    @staticmethod
+    def multi_choice_queryset(**kwargs):
+        setdefaults(kwargs, dict(
+            bulk__class=Field.multi_choice_queryset,
+            bulk__model=kwargs.get('model'),
+            query__class=Variable.multi_choice_queryset,
+            query__model=kwargs.get('model'),
+            cell__format=lambda value, **_: ', '.join(['%s' % x for x in value.all()]),
+        ))
+        return Column.choice(**kwargs)
+
     @staticmethod
     def choice(**kwargs):
         choices = kwargs['choices']
@@ -355,6 +367,51 @@ class Column(Frozen, ColumnBase):
         ))
         return Column(**kwargs)
 
+    @staticmethod
+    def date(**kwargs):
+        setdefaults(kwargs, dict(
+            query__gui__class=Field.date,
+            query__op_to_q_op=lambda op: {'=': 'exact', ':': 'contains'}.get(op) or Q_OP_BY_OP[op],
+            bulk__class=Field.date,
+        ))
+        return Column(**kwargs)
+
+    @staticmethod
+    def datetime(**kwargs):
+        setdefaults(kwargs, dict(
+            query__gui__class=Field.date,
+            query__op_to_q_op=lambda op: {'=': 'exact', ':': 'contains'}.get(op) or Q_OP_BY_OP[op],
+            bulk__class=Field.date,
+        ))
+        return Column(**kwargs)
+
+    @staticmethod
+    def email(**kwargs):
+        setdefaults(kwargs, dict(
+            bulk__class=Field.email,
+            # TODO: query__class=Variable.email,
+        ))
+        return Column(**kwargs)
+
+    @staticmethod
+    def from_model(model, field_name=None, model_field=None, **kwargs):
+        return member_from_model(
+            model=model,
+            factory_lookup=_column_factory_by_django_field_type,
+            field_name=field_name,
+            model_field=model_field,
+            defaults_factory=lambda model_field: {},
+            **kwargs)
+
+    @staticmethod
+    def expand_member(model, field_name=None, model_field=None, **kwargs):
+        return expand_member(
+            model=model,
+            factory_lookup=_column_factory_by_django_field_type,
+            field_name=field_name,
+            model_field=model_field,
+            **kwargs)
+
 
 class BoundColumn(ColumnBase):
 
@@ -367,6 +424,15 @@ class BoundColumn(ColumnBase):
     is_sorting = NamedStructField()
     """ :type: bool """
 
+    def __init__(self, **kwargs):
+        new_kwargs = setdefaults_path(
+            Struct(),
+            kwargs,
+            bulk__attr=kwargs.get('attr'),
+            query__attr=kwargs.get('attr'),
+        )
+        super(BoundColumn, self).__init__(**new_kwargs)
+
     def render_css_class(self):
         return render_class(self.attrs['class'])
 
@@ -469,7 +535,7 @@ class BoundCell(object):
         return "<%s column=%s row=%s>" % (self.__class__.__name__, self.bound_column.column, self.bound_row.row)  # pragma: no cover
 
 
-@declarative(Column, 'columns')
+@declarative(Column, 'columns_dict')
 @with_meta
 class Table(object):
 
@@ -500,7 +566,7 @@ class Table(object):
 
         model = None
 
-    def __init__(self, data, request=None, columns=None, **kwargs):
+    def __init__(self, data=None, request=None, columns=None, columns_dict=None, **kwargs):
         """
         :param data: a list of QuerySet of objects
         :param columns: (use this only when not using the declarative style) a list of Column objects
@@ -515,23 +581,29 @@ class Table(object):
         :param links__template:
 
         """
-        assert columns is not None, 'columns must be specified. It is only set to None to make linting tools not give false positives on the declarative style'
 
         self._has_prepared = False
 
+        if data is None:
+            assert 'model' in kwargs and kwargs['model'] is not None
+            data = kwargs['model'].objects.all()
+
         self.data = data
         self.request = request
 
         if isinstance(self.data, QuerySet):
             kwargs['model'] = data.model
 
-        if isinstance(columns, dict):
-            for name, column in columns.items():
+        def generate_columns():
+            for column in columns if columns is not None else []:
+                yield column
+            for name, column in columns_dict.items():
                 dict.__setitem__(column, 'name', name)
-            self.columns = list(columns.values())
-        else:
-            self.columns = columns
-            """:type : list of Column"""
+                yield column
+        self.columns = sort_after(list(generate_columns()))
+        """:type : list of Column"""
+
+        assert len(self.columns) > 0, 'columns must be specified. It is only set to None to make linting tools not give false positives on the declarative style'
 
         self.bound_columns = None
         self.shown_bound_columns = None
@@ -645,7 +717,8 @@ class Table(object):
 
             header_group[0].attrs['class']['first_column'] = True
 
-        header_groups[0].attrs['class']['first_column'] = True
+        if header_groups:
+            header_groups[0].attrs['class']['first_column'] = True
 
         for x in header_groups:
             if not isinstance(x.display_name, string_types):
@@ -677,36 +750,33 @@ class Table(object):
         self._prepare_evaluate_members()
         self._prepare_sorting()
         headers = self._prepare_headers()
-        self._prepare_auto_rowspan()
 
         if self.Meta.model:
-            def bulk(column):
-                bulk_kwargs = {
-                    'name': column.name,
-                    'attr': column.attr,
-                    'required': False,
-                    'empty_choice_tuple': (None, '', '---', True),
-                    'model': self.Meta.model,
-                    'class': Field.from_model,
-                }
-                bulk_kwargs.update(column.bulk)
-                if bulk_kwargs['class'] == Field.from_model:
-                    bulk_kwargs['field_name'] = column.attr
-                return bulk_kwargs.pop('class')(**bulk_kwargs)
 
-            def query(column):
-                query_kwargs = {
-                    'class': Variable,
-                    'name': column.name,
-                    'gui__label': column.display_name,
-                    'attr': column.attr,
-                    'model': column.table.Meta.model,
-                }
-                query_kwargs.update(column.query)
-                return query_kwargs.pop('class')(**query_kwargs)
+            def generate_variables():
+                for column in self.bound_columns:
+                    if column.query.show:
+                        query_kwargs = setdefaults_path(
+                            Struct(),
+                            column.query,
+                            dict(
+                                name=column.name,
+                                gui__label=column.display_name,
+                                attr=column.attr,
+                                model=column.table.Meta.model,
+                            ), {
+                                'class': Variable,
+                            }
+                        )
+                        yield query_kwargs.pop('class')(**query_kwargs)
+            variables = list(generate_variables())
 
-            self.query = Query(request=request, variables=[query(bound_column) for bound_column in self.bound_columns if bound_column.query.show], **self.query_kwargs)
-            self.query_form = self.query.form(request) if self.query.variables else None
+            self.query = Query(
+                request=request,
+                variables=variables,
+                **self.query_kwargs
+            )
+            self.query_form = self.query.form() if self.query.variables else None
 
             self.query_error = ''
             if self.query_form:
@@ -715,8 +785,34 @@ class Table(object):
                 except QueryException as e:
                     self.query_error = str(e)
 
-            bulk_fields = [bulk(bound_column) for bound_column in self.bound_columns if bound_column.bulk.show]
-            self.bulk_form = Form(data=request.POST, fields=bulk_fields, **self.bulk_kwargs) if bulk_fields else None
+            def generate_bulk_fields():
+                for column in self.bound_columns:
+                    if column.bulk.show:
+                        bulk_kwargs = setdefaults_path(
+                            Struct(),
+                            column.bulk,
+                            dict(
+                                name=column.name,
+                                attr=column.attr,
+                                required=False,
+                                empty_choice_tuple=(None, '', '---', True),
+                                model=self.Meta.model,
+                            ), {
+                                'class': Field.from_model,
+                            }
+                        )
+                        if bulk_kwargs['class'] == Field.from_model:
+                            bulk_kwargs['field_name'] = column.attr
+                        yield bulk_kwargs.pop('class')(**bulk_kwargs)
+            bulk_fields = list(generate_bulk_fields())
+
+            self.bulk_form = Form(
+                data=request.POST,
+                fields=bulk_fields,
+                endpoint_dispatch_prefix='bulk',
+                **self.bulk_kwargs) if bulk_fields else None
+
+        self._prepare_auto_rowspan()
 
         return headers, self.header_levels
 
@@ -731,6 +827,39 @@ class Table(object):
     def render_tbody(self):
         return '\n'.join([bound_row.render() for bound_row in self.bound_rows()])
 
+    @staticmethod
+    def columns_from_model(**kwargs):
+        kwargs = collect_namespaces(kwargs)
+        kwargs['db_field'] = kwargs.pop('column', {})
+        return create_members_from_model(default_factory=Column.from_model, **kwargs)
+
+    @staticmethod
+    def from_model(data, model, instance=None, include=None, exclude=None, extra_fields=None, post_validation=None, **kwargs):
+        """
+        Create an entire form based on the fields of a model. To override a field parameter send keyword arguments in the form
+        of "the_name_of_the_field__param". For example:
+
+        .. code:: python
+
+            class Foo(Model):
+                foo = IntegerField()
+
+            Table.from_model(data=request.GET, model=Foo, field__foo__help_text='Overridden help text')
+
+        :param include: fields to include. Defaults to all
+        :param exclude: fields to exclude. Defaults to none (except that AutoField is always excluded!)
+
+        """
+        kwargs = collect_namespaces(kwargs)
+        columns = Table.columns_from_model(model=model, include=include, exclude=exclude, extra=extra_fields, column=kwargs.pop('column', {}))
+        return Table(data=data, model=model, instance=instance, columns=columns, post_validation=post_validation, **kwargs)
+
+    def endpoint_dispatch(self, key, value):
+        if key.startswith('query__'):
+            return self.query.endpoint_dispatch(key=key[len('query__'):], value=value)
+        if key.startswith('bulk__'):
+            return self.bulk.endpoint_dispatch(key=key[len('bulk__'):], value=value)
+
 
 class Link(Struct):
     """
@@ -761,6 +890,8 @@ def table_context(request,
     if extra_context is None:  # pragma: no cover
         extra_context = {}
 
+    assert table.data is not None
+
     grouped_links = {}
     if links is not None:
         links = evaluate_recursive(links, table=table)
@@ -834,7 +965,7 @@ def set_display_none(rowspan_by_row):
 
 
 def render_table(request,
-                 table,
+                 table=None,
                  links=None,
                  context=None,
                  template_name='tri_table/list.html',
@@ -844,7 +975,8 @@ def render_table(request,
                  context_processors=None,
                  paginator=None,
                  show_hits=False,
-                 hit_label='Items'):
+                 hit_label='Items',
+                 **kwargs):
     """
     Render a table. This automatically handles pagination, sorting, filtering and bulk operations.
 
@@ -861,7 +993,22 @@ def render_table(request,
     if not context:
         context = {}
 
+    kwargs = collect_namespaces(kwargs)
+
+    if table is None or isinstance(table, dict):
+        table_kwargs = table if isinstance(table, dict) else kwargs.pop('table', {})
+        if 'model' not in table_kwargs:
+            table_kwargs['model'] = table_kwargs['data'].model
+        table = Table.from_model(**table_kwargs)
+
     table.prepare(request)
+    assert isinstance(table, Table)
+
+    for key, value in request.GET.items():
+        if key.startswith('__'):
+            data = table.endpoint_dispatch(key=key[2:], value=value)
+            if data:
+                return HttpResponse(json.dumps(data), content_type='application/json')
 
     context['bulk_form'] = table.bulk_form
     context['query_form'] = table.query_form
@@ -893,6 +1040,10 @@ def render_table(request,
     if not table.data and blank_on_empty:  # pragma: no cover
         return ''
 
+    if table.query_form and not table.query_form.is_valid():
+        table.data = None
+        context['invalid_form_message'] = mark_safe('<i class="fa fa-meh-o fa-5x" aria-hidden="true"></i>')
+
     return get_template(template_name).render(context)
 
 
@@ -904,3 +1055,7 @@ def render_table_to_response(*args, **kwargs):
     if isinstance(response, HttpResponse):  # pragma: no cover
         return response
     return HttpResponse(response)
+
+from .db_compat import setup, _column_factory_by_django_field_type  # noqa
+
+setup()
